from datetime import date
from typing import Optional

from django.db.models import Model
from django.db.models.fields import CharField, FloatField, BooleanField, DateField, DateTimeField
from django.db.models.fields.related import ForeignKey
from django.db.models.deletion import PROTECT, CASCADE
from django.db.models.enums import TextChoices
from django.core.exceptions import ObjectDoesNotExist

from django.utils.translation import gettext_lazy as _


class AbstractBaseModel(Model):

    date_created = DateTimeField(
        auto_now_add=True,
        verbose_name=_("Time of creation")
    )
    date_modified = DateTimeField(
        auto_now=True,
        verbose_name=_("Time of last modification")
    )

    class Meta:
        abstract = True


class FinancialPosition(AbstractBaseModel):

    class FinStmt(TextChoices):
        BS = 'BS', _("Balance Sheet")
        IS = 'IS', _("Income Statement")
        CF = 'CF', _("Cash Flow Statement")

    financial_statement = CharField(
        max_length=2,
        db_index=True,
        choices=FinStmt.choices,
        verbose_name=_("Financial statement")
    )
    name = CharField(
        max_length=255,
        db_index=True,
        verbose_name=_("Position name")
    )
    parent = ForeignKey(
        to='self',
        on_delete=PROTECT,
        null=True,
        blank=True,
        verbose_name=_("Parent position")
    )

    def __str__(self) -> str:
        return f"{self.financial_statement} - {self.name}"


class ReportingPeriod(AbstractBaseModel):

    date_end = DateField(
        db_index=True,
        verbose_name=_("Period end date")
    )
    date_start = DateField(
        db_index=True,
        verbose_name=_("Period start date")
    )
    verified = BooleanField(
        default=False,
        verbose_name=_("Exact dates verified")
    )

    def __str__(self) -> str:
        date_fmt = '%b %Y'
        assert isinstance(self.date_end, date) and isinstance(self.date_start, date)
        return f"{'Quarter' if self.is_quarter else 'Year'}: " \
               f"{self.date_start.strftime(date_fmt)} - {self.date_end.strftime(date_fmt)}"

    @property
    def is_quarter(self) -> bool:
        assert isinstance(self.date_end, date) and isinstance(self.date_start, date)
        delta = self.date_end - self.date_start
        return delta.days < 100

    class Meta:
        ordering = ['-date_end', 'date_start']


class Company(AbstractBaseModel):

    symbol = CharField(
        max_length=16,
        unique=True,
        verbose_name=_("Stock ticker symbol")
    )

    def __str__(self) -> str:
        return f"{self.name if self.name else '[~~~ NO NAME ~~~]'} ({self.symbol})"

    @property
    def name(self) -> Optional[str]:
        if self.current_name_obj is not None:
            return self.current_name_obj.name

    @property
    def current_name_obj(self) -> Optional['CompanyName']:
        try:
            return self.company_names.latest()
        except ObjectDoesNotExist:
            return None

    class Meta:
        verbose_name_plural = _("Companies")


class CompanyName(AbstractBaseModel):

    company = ForeignKey(
        to=Company,
        on_delete=CASCADE,
        related_name='company_names',
        related_query_name='company_name',
    )
    name = CharField(
        max_length=1024,
        db_index=True,
        verbose_name=_("Name")
    )
    name_since = DateField(
        db_index=True,
        null=True,
        blank=True,
        verbose_name=_("Has this name since")
    )

    def __str__(self) -> str:
        return self.name

    class Meta:
        get_latest_by = ['name_since', 'date_modified']


class Figure(AbstractBaseModel):

    position = ForeignKey(
        to=FinancialPosition,
        on_delete=PROTECT
    )
    period = ForeignKey(
        to=ReportingPeriod,
        on_delete=PROTECT
    )
    company = ForeignKey(
        to=Company,
        on_delete=PROTECT
    )
    value = FloatField(
        verbose_name=_("Figure on financial statement")
    )

    class Meta:
        ordering = ['company', 'period', 'position']
